#include "maboutdialog.h"
#include "ui_maboutdialog.h"

// Qt
#include <QDesktopServices>
#include <QGuiApplication>
#include <QJsonDocument>
#include <QJsonObject>
#include <QNetworkAccessManager>
#include <QNetworkReply>
#include <QPropertyAnimation>
#include <QUrl>


#include <MHelpBrowser>

class MAboutDialog::Private
{
public:
    Private() :
        ui(new Ui_MAboutDialog),
        author(),
        applicationName(),
        applicationVersion(),
        homepage(),
        authorsHelpPage(),
        licenseHelpPage(),
        updateUrl(),
        pixmap(),
        qnam(),
        reply(nullptr)
    {}

    ~Private()
    {
        delete(ui);
    }

    Ui_MAboutDialog *ui;
    QString author;
    QString applicationName;
    QString applicationVersion;
    QString homepage;
    QString authorsHelpPage;
    QString licenseHelpPage;
    QUrl updateUrl;
    QPixmap pixmap;
    QNetworkAccessManager qnam;
    QNetworkReply *reply;
};

MAboutDialog::MAboutDialog(QWidget *parent) :
    QDialog(parent),
    d_ptr(new Private)
{
    d_ptr->ui->setupUi(this);

    setWindowFlags(windowFlags() & ~Qt::WindowContextHelpButtonHint);

    connect(this, &MAboutDialog::authorChanged, [this](const QString &author) {
        d_ptr->ui->createdBy_label->setText(tr("Created by %1").arg(author));
    });

    connect(this, &MAboutDialog::applicationNameChanged, [this](const QString applicationName) {
        d_ptr->ui->appName_label->setText(QString(QLatin1String("<b>%1</b>")).arg(applicationName));
    });

    connect(this, &MAboutDialog::applicationVersionChanged, [this](const QString applicationVersion) {
        d_ptr->ui->appVersion_label->setText(tr("Version %1").arg(applicationVersion));
    });

    connect(this, &MAboutDialog::homepageChanged, this, &MAboutDialog::updateLinks);
    connect(this, &MAboutDialog::authorsHelpPageChanged, this, &MAboutDialog::updateLinks);
    connect(this, &MAboutDialog::licenseHelpPageChanged, this, &MAboutDialog::updateLinks);

    connect(this, &MAboutDialog::pixmapChanged, [this](const QPixmap &pixmap) {
        if (!pixmap.isNull()) {
            d_ptr->ui->icon_label->setPixmap(pixmap.scaledToWidth(64));
        }
    });

    connect(this, &MAboutDialog::updateUrlChanged, [this](const QUrl &url) {
        if (!url.isEmpty()) {

            d_ptr->ui->update_stackedWidget->setVisible(true);
            d_ptr->ui->update_line->setVisible(true);
        } else {
            d_ptr->ui->update_stackedWidget->setVisible(false);
            d_ptr->ui->update_line->setVisible(false);
        }
    });

    connect(d_ptr->ui->links_label, &QLabel::linkActivated, [=](const QString &link) {
        if (link.startsWith("http")) {
            QDesktopServices::openUrl(link);
        } else {
            this->hide();
            MHelpBrowser::open(link);
        }
    });

    connect(d_ptr->ui->update_pushButton, &QPushButton::clicked, [this]() {
        d_ptr->ui->update_stackedWidget->setCurrentIndex(1);

        d_ptr->reply = d_ptr->qnam.get(QNetworkRequest(updateUrl()));
        connect(d_ptr->reply, &QNetworkReply::finished, [this]() {
            if (d_ptr->reply->error()) {
                d_ptr->ui->update_stackedWidget->setCurrentIndex(4);
            } else {
                QJsonDocument doc = QJsonDocument::fromJson(d_ptr->reply->readAll());
                double latest = doc.object().value("latest").toDouble();
                double current = qApp->applicationVersion().toDouble();
                if (latest <= current) {
                    d_ptr->ui->update_stackedWidget->setCurrentIndex(2);
                } else if (latest > current) {
                    d_ptr->ui->update_stackedWidget->setCurrentIndex(3);
                }
            }
            d_ptr->reply->deleteLater();
        });
    });

    setPixmap(qApp->windowIcon());

    setAuthor(QCoreApplication::organizationName());
    setApplicationName(QCoreApplication::applicationName());
    setApplicationVersion(QCoreApplication::applicationVersion());
    emit updateUrlChanged(QUrl());

    adjustSize();
}

MAboutDialog::~MAboutDialog()
{

}




void MAboutDialog::adjustSize()
{
    setMinimumSize(0,0);
    setMaximumSize(0,0);
    QDialog::adjustSize();
}

void MAboutDialog::updateLinks()
{
    QString homepageLabel = tr("Homepage");
    QString licenseLabel = tr("License");
    QString authorsLabel = tr("Authors");
    QStringList entries;

    if (!homepage().isEmpty()) {
        entries << QString(QLatin1String("<a href=\"%1\">%2</a>")).arg(homepage()).arg(homepageLabel);
    }

    if (!licenseHelpPage().isEmpty()) {
        entries << QString(QLatin1String("<a href=\"%1\">%2</a>")).arg(licenseHelpPage()).arg(licenseLabel);
    }

    if (!authorsHelpPage().isEmpty()) {
        entries << QString(QLatin1String("<a href=\"%1\">%2</a>")).arg(authorsHelpPage()).arg(authorsLabel);
    }

    d_ptr->ui->links_label->setText(entries.join(" | "));
    adjustSize();
}


QString &MAboutDialog::author() const { return d_ptr->author; }
void MAboutDialog::setAuthor(const QString &author)
{
    if (d_ptr->author == author)
        return;

    d_ptr->author = author;
    emit authorChanged(author);
}

QString &MAboutDialog::applicationName() const { return d_ptr->applicationName; }
void MAboutDialog::setApplicationName(const QString &applicationName)
{
    if (d_ptr->applicationName == applicationName)
        return;

    d_ptr->applicationName = applicationName;
    emit applicationNameChanged(applicationName);
}

QString &MAboutDialog::applicationVersion() const { return d_ptr->applicationVersion; }
void MAboutDialog::setApplicationVersion(const QString &applicationVersion)
{
    if (d_ptr->applicationVersion == applicationVersion)
        return;

    d_ptr->applicationVersion = applicationVersion;
    emit applicationVersionChanged(applicationVersion);
}

QString &MAboutDialog::homepage() const { return d_ptr->homepage; }
void MAboutDialog::setHomepage(const QString &homepage)
{
    if (d_ptr->homepage == homepage)
        return;

    d_ptr->homepage = homepage;
    emit homepageChanged(homepage);
}

QString &MAboutDialog::authorsHelpPage() const { return d_ptr->authorsHelpPage; }
void MAboutDialog::setAuthorsHelpPage(const QString &authorsHelpPage)
{
    if (d_ptr->authorsHelpPage == authorsHelpPage)
        return;

    d_ptr->authorsHelpPage = authorsHelpPage;
    emit authorsHelpPageChanged(authorsHelpPage);
}

QString &MAboutDialog::licenseHelpPage() const { return d_ptr->licenseHelpPage; }
void MAboutDialog::setLicenseHelpPage(const QString &licenseHelpPage)
{
    if (d_ptr->licenseHelpPage == licenseHelpPage)
        return;

    d_ptr->licenseHelpPage = licenseHelpPage;
    emit licenseHelpPageChanged(licenseHelpPage);
}

QUrl &MAboutDialog::updateUrl() const { return d_ptr->updateUrl; }
void MAboutDialog::setUpdateUrl(const QUrl &updateUrl)
{
    if (d_ptr->updateUrl == updateUrl)
        return;

    d_ptr->updateUrl = updateUrl;
    emit updateUrlChanged(updateUrl);
}

QPixmap &MAboutDialog::pixmap() const { return d_ptr->pixmap; }
void MAboutDialog::setPixmap(const QPixmap &pixmap)
{
    if (d_ptr->pixmap == pixmap)
        return;

    d_ptr->pixmap = pixmap;
    emit pixmapChanged(pixmap);
}

void MAboutDialog::setPixmap(const QIcon &icon)
{
    if (!icon.isNull()) {
        setPixmap(icon.pixmap(64, 64));
    } else {
        setPixmap(icon);
    }
}
