#include "stash.h"

// Qt
#include <QCoreApplication>
#include <QDebug>
#include <QDir>
#include <QFile>
#include <QFileInfo>
#include <QMutex>
#include <QProcess>
#include <QSettings>
#include <QTextStream>
#include <QThread>

#include "keystore.h"
#include "encfs.h"
#include "helpers.h"
#include "settings.h"
#include "stashmanager.h"

#include <mntent.h>

QMutex stash_env_mutex;
int stash_instances = 0;

Stash::Stash(QObject *parent) :
    QObject(parent),
    m_encfs(new Encfs),
    m_isMounted(false),
    m_isAvailable(false),
    m_isInitialized(false),
    m_passwordInKeyring(false),
    m_action(Action_Idle),
    m_createFlags(0)
{
    //stash_instances++; qDebug() << "stash_instances:" << stash_instances;
    buildConnections();

    QSettings settings;
    settings.beginGroup(QLatin1String("StashDefaults"));
    load(settings);
}


Stash::Stash(const QSettings &settings, QObject *parent) :
    QObject(parent),
    m_encfs(new Encfs),
    m_isMounted(false),
    m_isAvailable(false),
    m_isInitialized(false),
    m_passwordInKeyring(false),
    m_action(Action_Idle),
    m_createFlags(0)
{
    //stash_instances++; qDebug() << "stash_instances:" << stash_instances;
    buildConnections();

    load(settings);
}

void Stash::buildConnections()
{
    // shortName
    connect(this, &Stash::mountDirChanged, [this](const QString &mountDir) {
        setShortName(QFileInfo(mountDir).fileName());
    });

    connect(this, &Stash::__deleteFinished, this, &Stash::deleteFinished, Qt::QueuedConnection);

    connect(&m_encfs, &Encfs::unmountFinished, [this](bool success, const QString &message) {
        refresh();

        //qDebug() << "stash.cpp: unmountFinished: " << success << message;
        if (success) {
            if (!persistentMountDirectory()) {
                QDir dir(mountDir());
                dir.rmdir(dir.absolutePath());
            }
        }

        if (success && !postUnmountCommand().isEmpty()) {
            if (!execCommandLine(postUnmountCommand())) {
                setBusy(Action_Idle);
                emit unmountFinished(false, tr("The stash was unmounted, but the custom command to run"
                                               "after that returned with a value other than 0."));
                return;
            }
        }

        setBusy(Action_Idle);
        emit unmountFinished(success, message);
    });

    connect(&m_encfs, &Encfs::mountFinished, [this](bool success, const QString &message) {
        refresh();

        //qDebug() << "stash.cpp: mountFinished: " << success << message;

        if (success && !m_passwordInKeyring && savePassword() && !message.isEmpty()) {
            KeyStore *keyStore = KeyStore::newKeyStore();
            connect(keyStore, &KeyStore::storePasswordFinished, keyStore, &QObject::deleteLater);
            keyStore->storePassword(sharedPointer(), message);
        }

        if (!success) {
            if (!persistentMountDirectory()) {
                QDir dir(mountDir());
                dir.rmdir(dir.absolutePath());
            }
        }

        if (success && !postMountCommand().isEmpty()) {
            if (!execCommandLine(postMountCommand())) {
                setBusy(Action_Idle);
                emit unmountFinished(false, tr("The stash was mounted, but the custom command to run"
                                               "after that returned with a value other than 0."));
                return;
            }
        }

        setBusy(Action_Idle);
        if (success) {
            emit mountFinished(success, QLatin1String());
        } else {
            emit mountFinished(success, message);
        }
    });

    connect(&m_encfs, &Encfs::changePasswordFinished, [this](bool success, const QString &message) {
        refresh();
        setBusy(Action_Idle);
        //qDebug() << "stash.cpp: changePasswordFinished: " << success << message;

        if (success && savePassword()) {
            KeyStore *keyStore = KeyStore::newKeyStore();
            connect(keyStore, &KeyStore::storePasswordFinished, keyStore, &QObject::deleteLater);
            keyStore->storePassword(sharedPointer(), message);
        }

        if (success) {
            emit changePasswordFinished(success, QLatin1String());
        } else {
            emit changePasswordFinished(success, message);
        }
    });

    connect(&m_encfs, &Encfs::createFinished, [this](bool success, const QString &message) {
        refresh();
        setBusy(Action_Idle);
        //qDebug() << "stash.cpp: createFinished: " << success << message;

        if (!success) {
            QDir _stashDir(stashDir());
            if (_stashDir.isEmpty(QDir::NoDotAndDotDot)) {
                _stashDir.rmdir(_stashDir.absolutePath());
            }
        }

        if (success && savePassword()) {
            KeyStore *keyStore = KeyStore::newKeyStore();
            connect(keyStore, &KeyStore::storePasswordFinished, keyStore, &QObject::deleteLater);
            keyStore->storePassword(sharedPointer(), message);
        }

        if (!success) {
            if (!persistentMountDirectory()) {
                QDir dir(mountDir());
                dir.rmdir(dir.absolutePath());
            }
        }

        if (success) {
            setIsInitialized(true);
            emit createFinished(success, QLatin1String());
        } else {
            emit createFinished(success, message);
        }
    });

    connect(this, &Stash::mountDirChanged, [this]() {
        refresh();
    });
    connect(this, &Stash::stashDirChanged, [this]() {
        refresh();
    });
    connect(this, &Stash::encfsConfigFileChanged, [this]() {
        refresh();
    });

    connect(this, &Stash::savePasswordChanged, [this](bool savePassword) {
        if (!isInitialized()) { return; }

        if (!savePassword) {
            KeyStore *keyStore = KeyStore::newKeyStore();
            connect(keyStore, &KeyStore::storePasswordFinished, keyStore, &QObject::deleteLater);
            keyStore->deletePassword(sharedPointer());
            m_passwordInKeyring = false;
        }
    });


    // Persistent mount directory -----------------------------------------------------------------
    if (isInitialized()) {
        if (persistentMountDirectory()) {
            QDir(this->mountDir()).mkpath(QLatin1String("."));
        } else {
            if (!isMounted()) {
                QDir dir(this->mountDir());
                if (dir.exists()) {
                    dir.rmdir(dir.absolutePath());
                }
            }
        }
    }
    connect(this, &Stash::persistentMountDirectoryChanged, [this](bool persistentMountDirectory) {
        if (!isInitialized()) { return; }

        if (persistentMountDirectory) {
            QDir(this->mountDir()).mkpath(QLatin1String("."));
        } else {
            QDir dir(this->mountDir());
            if (dir.exists()) {
                dir.rmdir(dir.absolutePath());
            }
        }
    });
    connect(this, &Stash::isMountedChanged, [this](bool isMounted) {
        if (!isInitialized()) { return; }

        if (!isMounted && !persistentMountDirectory()) {
            QDir dir(this->mountDir());
            if (dir.exists()) {
                dir.rmdir(dir.absolutePath());
            }
        }
    });

    // Open in file manager -----------------------------------------------------------------------
    connect(this, &Stash::isMountedChanged, [this](bool isMounted) {
        if (!isInitialized()) { return; }

        if (isMounted && openFileManager()) {
            execCommandLine(Settings::instance().fileManagerCommand(), true);
        }
    });

    // Mount when available -----------------------------------------------------------------------
    connect(this, &Stash::isAvailableChanged, [this](bool isAvailable) {
        if (!isInitialized()) { return; }

        if (isAvailable && !isMounted() && mountWhenAvailable()) {
            StashManager::instance().mount(sharedPointer());
        }
    });

    // Unmount stash when unavailable (maybe removed drive)----------------------------------------
    connect(this, &Stash::isAvailableChanged, [this](bool isAvailable) {
        if (!isInitialized()) { return; }

        if (!isAvailable && isMounted()) {
            unmount();
        }
    });
}

Encfs &Stash::encfs()
{
    return m_encfs;
}

Stash::~Stash()
{
    //stash_instances--; qDebug() << "stash_instances:" << stash_instances << "|" << mountDir();
}


void Stash::mount(const QString &password)
{
    if (isBusy()) {
        emit mountFinished(false, tr("The stash is currently busy"));
        return;
    }
    if (!isAvailable()) {
        emit mountFinished(false, tr("The stash is not available"));
        return;
    }
    if (isMounted()) {
        emit mountFinished(false, tr("The stash is already mounted"));
        return;
    }

    setBusy(Action_Mounting, tr("Mounting %1 …").arg(shortName()));

    if (!preMountCommand().isEmpty()) {
        if (!execCommandLine(preMountCommand())) {
            setBusy(Action_Idle);
            emit mountFinished(false, tr("The custom command to run before mounting "
                                           "returned with a value other than 0."));
            return;
        }
    }

    if (password.isEmpty()) {
        if (savePassword()) {
            // Look up password in keyring
            KeyStore *keyStore = KeyStore::newKeyStore();
            connect(keyStore, &KeyStore::retrievePasswordFinished, [this, keyStore](bool success, const QString &passwordOrMessage) {
                keyStore->deleteLater();
                if (success) {
                    //qDebug() << "from keyring";
                    m_passwordInKeyring = true;
                    m_encfs.mount(sharedPointer(), passwordOrMessage);
                } else {
                    //qDebug() << "failed";
                    m_passwordInKeyring = false;
                    setBusy(Action_Idle);
                    emit mountFinished(false, passwordOrMessage);
                }
            });
            keyStore->retrievePassword(sharedPointer());
        } else {
            // open password dialog
            setBusy(Action_Idle);
            emit mountFinished(false, tr("Please enter the password"));
        }
    } else {
        m_encfs.mount(sharedPointer(), password);
    }
}

void Stash::unmount()
{
    if (isBusy() || !isMounted()) {
        emit unmountFinished(false, tr("The stash is currently busy"));
        return;
    }

    setBusy(Action_Unmounting, tr("Unmounting %1 …").arg(shortName()));

    if (!preUnmountCommand().isEmpty()) {
        if (!execCommandLine(preUnmountCommand())) {
            setBusy(Action_Idle);
            emit unmountFinished(false, tr("The custom command to run before mounting "
                                           "returned with a value other than 0."));
            return;
        }
    }

    m_encfs.unmount(sharedPointer());
}

void Stash::create(const QString &password, Encfs::CreateStashFlags flags)
{
    Q_UNUSED(flags)

    if (isBusy()) {
        emit createFinished(false, tr("The stash is currently busy"));
        return;
    }
    setBusy(Action_Creating, tr("Creating %1 …").arg(shortName()));
    m_encfs.create(sharedPointer(), password, m_createFlags);
}

void Stash::changePassword(const QString &oldPassword, const QString &newPassword)
{
    if (isBusy()) {
        emit changePasswordFinished(false, tr("The stash is currently busy"));
        return;
    }
    setBusy(Action_ChangingPassword, tr("Changing password for %1 …").arg(shortName()));
    m_encfs.changePassword(sharedPointer(), oldPassword, newPassword);
}

QString Stash::info() const
{
   return Encfs().info(sharedPointer());
}

void Stash::cancel()
{
    m_encfs.cancel();
}

void Stash::delete_()
{
    if (isMounted()) {
        emit deleteFinished(false, tr("The stash is mounted"));
        return;
    }

    if (!isAvailable()) {
        emit deleteFinished(false, tr("The stash is not available"));
        return;
    }

    if (isBusy()) {
        emit deleteFinished(false, tr("The stash is currently busy"));
        return;
    }

    setBusy(Action_Deleting, tr("Deleting %1 …").arg(shortName()));

    QThread *thread = QThread::create([this]() {
        QProcess process;
        process.setArguments(QStringList() << QLatin1String("-rf") << stashDir());
        process.setProgram(QLatin1String("rm"));
        process.start();
        process.waitForFinished();

        setBusy(Action_Idle);
        refresh();
        if (isAvailable()) {
            emit __deleteFinished(false, tr("Failed to delete the stash"));
        } else {
            emit __deleteFinished(true);
        }

    });
    thread->start();
}

bool Stash::execCommandLine(const QString &commandLine, bool startDetached)
{
    QProcess process;
    process.setProcessChannelMode(QProcess::ForwardedChannels);
    bool ret = true;

    QTextStream out(stdout);
    out << "Command: " << commandLine << endl;
    if (startDetached) {
        // setenv doesn't work work with startDetached(), seems to be a Qt bug, tested with Qt 5.11
        stash_env_mutex.lock();
        qputenv("MOUNT_DIR", mountDir().toUtf8());
        qputenv("STASH_DIR", stashDir().toUtf8());
        process.startDetached(QLatin1String("bash"), QStringList() << QLatin1String("-c") << commandLine);
        qunsetenv("MOUNT_DIR");
        qunsetenv("STASH_DIR");
        stash_env_mutex.unlock();
    } else {
        QProcessEnvironment env = QProcessEnvironment::systemEnvironment();
        env.insert(QLatin1String("MOUNT_DIR"), mountDir().toUtf8());
        env.insert(QLatin1String("STASH_DIR"), stashDir().toUtf8());
        process.setProcessEnvironment(env);
        process.start(QLatin1String("bash"), QStringList() << QLatin1String("-c") << commandLine);
        process.waitForFinished();
        ret = process.exitCode() == 0;
    }

    return ret;
}

void Stash::refresh()
{
    bool mounted = false;
    if (!m_mountDir.isEmpty()) {
        FILE *f = setmntent("/etc/mtab", "r");
        for (;;) {
            struct mntent *m = getmntent(f);
            if (!m) break;
            if (m_mountDir == m->mnt_dir) {
                mounted = true;
                break;
            }
        }
        endmntent(f);
    }

    // stash directory and encfs config file have to exist
    bool available = false;
    if (!stashDir().isEmpty() && QDir(stashDir()).exists()) {
        if (encfsConfigFile().isEmpty()) {
            if (QFile(stashDir() + QLatin1String("/.encfs6.xml")).exists()) {
                available = true;
            }
        } else {
            if (QFile(encfsConfigFile()).exists() && QFileInfo(encfsConfigFile()).isFile()) {
                available = true;
            }
        }
    }

    setIsAvailable(available);
    setIsMounted(mounted);
}





/* Used the script in in the folder DevTools to autogenerate these
   Also feed directories through the helpers to make them relative or absolute */
// Saved Properties ===============================================================================
void Stash::load(const QSettings &settings)
{
    setStashDir(settings.value(QLatin1String("stashDir"), QLatin1String("")).toString());
    setMountDir(settings.value(QLatin1String("mountDir"), QLatin1String("")).toString());
    setMountOnStartup(settings.value(QLatin1String("mountOnStartup"), true).toBool());
    setSavePassword(settings.value(QLatin1String("savePassword"), true).toBool());
    setStashGroup(settings.value(QLatin1String("stashGroup"), QLatin1String("")).toString());
    setShowInTray(settings.value(QLatin1String("showInTray"), true).toBool());
    setMountWhenAvailable(settings.value(QLatin1String("mountWhenAvailable"), true).toBool());
    setAllowOtherUsers(settings.value(QLatin1String("allowOtherUsers"), false).toBool());
    setPersistentMountDirectory(settings.value(QLatin1String("persistentMountDirectory"), false).toBool());
    setOpenFileManager(settings.value(QLatin1String("openFileManager"), false).toBool());
    setUnmountIdle(settings.value(QLatin1String("unmountIdle"), false).toBool());
    setUnmountIdleTime(settings.value(QLatin1String("unmountIdleTime"), 5).toInt());
    setEncfsConfigFile(settings.value(QLatin1String("encfsConfigFile"), QLatin1String("")).toString());
    setAdditionalMountOptions(settings.value(QLatin1String("additionalMountOptions"), QLatin1String("")).toString());
    setPreMountCommand(settings.value(QLatin1String("preMountCommand"), QLatin1String("")).toString());
    setPostMountCommand(settings.value(QLatin1String("postMountCommand"), QLatin1String("")).toString());
    setPreUnmountCommand(settings.value(QLatin1String("preUnmountCommand"), QLatin1String("")).toString());
    setPostUnmountCommand(settings.value(QLatin1String("postUnmountCommand"), QLatin1String("")).toString());
    setKeyringName(settings.value(QLatin1String("keyringName"), QLatin1String("")).toString());
    setAllowNonEmptyMountDirectory(settings.value(QLatin1String("allowNonEmptyMountDirectory"), false).toBool());
    setAllowRootAccess(settings.value(QLatin1String("allowRootAccess"), false).toBool());
    setTrayIconIndicateMounted(settings.value(QLatin1String("trayIconIndicateMounted"), false).toBool());
}

void Stash::save(QSettings &settings)
{
    if (stashDir() != QLatin1String("")) { settings.setValue(QLatin1String("stashDir"), Helpers::pathToRelative(stashDir())); }
    if (mountDir() != QLatin1String("")) { settings.setValue(QLatin1String("mountDir"), Helpers::pathToRelative(mountDir())); }
    if (mountOnStartup() != true) { settings.setValue(QLatin1String("mountOnStartup"), mountOnStartup()); }
    if (savePassword() != true) { settings.setValue(QLatin1String("savePassword"), savePassword()); }
    if (stashGroup() != QLatin1String("")) { settings.setValue(QLatin1String("stashGroup"), stashGroup()); }
    if (showInTray() != true) { settings.setValue(QLatin1String("showInTray"), showInTray()); }
    if (mountWhenAvailable() != true) { settings.setValue(QLatin1String("mountWhenAvailable"), mountWhenAvailable()); }
    if (allowOtherUsers() != false) { settings.setValue(QLatin1String("allowOtherUsers"), allowOtherUsers()); }
    if (persistentMountDirectory() != false) { settings.setValue(QLatin1String("persistentMountDirectory"), persistentMountDirectory()); }
    if (openFileManager() != false) { settings.setValue(QLatin1String("openFileManager"), openFileManager()); }
    if (unmountIdle() != false) { settings.setValue(QLatin1String("unmountIdle"), unmountIdle()); }
    if (unmountIdleTime() != 5) { settings.setValue(QLatin1String("unmountIdleTime"), unmountIdleTime()); }
    if (encfsConfigFile() != QLatin1String("")) { settings.setValue(QLatin1String("encfsConfigFile"), Helpers::pathToRelative(encfsConfigFile())); }
    if (additionalMountOptions() != QLatin1String("")) { settings.setValue(QLatin1String("additionalMountOptions"), additionalMountOptions()); }
    if (preMountCommand() != QLatin1String("")) { settings.setValue(QLatin1String("preMountCommand"), preMountCommand()); }
    if (postMountCommand() != QLatin1String("")) { settings.setValue(QLatin1String("postMountCommand"), postMountCommand()); }
    if (preUnmountCommand() != QLatin1String("")) { settings.setValue(QLatin1String("preUnmountCommand"), preUnmountCommand()); }
    if (postUnmountCommand() != QLatin1String("")) { settings.setValue(QLatin1String("postUnmountCommand"), postUnmountCommand()); }
    if (keyringName() != QLatin1String("")) { settings.setValue(QLatin1String("keyringName"), keyringName()); }
    if (allowNonEmptyMountDirectory() != false) { settings.setValue(QLatin1String("allowNonEmptyMountDirectory"), allowNonEmptyMountDirectory()); }
    if (allowRootAccess() != false) { settings.setValue(QLatin1String("allowRootAccess"), allowRootAccess()); }
    if (trayIconIndicateMounted() != false) { settings.setValue(QLatin1String("trayIconIndicateMounted"), trayIconIndicateMounted()); }
}

QString Stash::stashDir() const { return m_stashDir; }
void Stash::setStashDir (const QString &stashDir) {
    QString dir = Helpers::pathToAbsolute(stashDir);
    if (m_stashDir == dir)
        return;

    m_stashDir = dir;
    emit stashDirChanged(dir);
}

QString Stash::mountDir() const { return m_mountDir; }
void Stash::setMountDir (const QString &mountDir) {
    QString dir = Helpers::pathToAbsolute(mountDir);
    if (m_mountDir == dir)
        return;

    m_mountDir = dir;
    emit mountDirChanged(dir);
}

bool Stash::mountOnStartup() const { return m_mountOnStartup; }
void Stash::setMountOnStartup (bool mountOnStartup) {
    if (m_mountOnStartup == mountOnStartup)
        return;

    m_mountOnStartup = mountOnStartup;
    emit mountOnStartupChanged(mountOnStartup);
}

bool Stash::savePassword() const { return m_savePassword; }
void Stash::setSavePassword (bool savePassword) {
    if (m_savePassword == savePassword)
        return;

    m_savePassword = savePassword;
    emit savePasswordChanged(savePassword);
}

QString Stash::stashGroup() const { return m_stashGroup; }
void Stash::setStashGroup (const QString &stashGroup) {
    if (m_stashGroup == stashGroup)
        return;

    m_stashGroup = stashGroup;
    emit stashGroupChanged(stashGroup);
}

bool Stash::showInTray() const { return m_showInTray; }
void Stash::setShowInTray (bool showInTray) {
    if (m_showInTray == showInTray)
        return;

    m_showInTray = showInTray;
    emit showInTrayChanged(showInTray);
}

bool Stash::mountWhenAvailable() const { return m_mountWhenAvailable; }
void Stash::setMountWhenAvailable (bool mountWhenAvailable) {
    if (m_mountWhenAvailable == mountWhenAvailable)
        return;

    m_mountWhenAvailable = mountWhenAvailable;
    emit mountWhenAvailableChanged(mountWhenAvailable);
}

bool Stash::allowOtherUsers() const { return m_allowOtherUsers; }
void Stash::setAllowOtherUsers (bool allowOtherUsers) {
    if (m_allowOtherUsers == allowOtherUsers)
        return;

    m_allowOtherUsers = allowOtherUsers;
    emit allowOtherUsersChanged(allowOtherUsers);
}

bool Stash::persistentMountDirectory() const { return m_persistentMountDirectory; }
void Stash::setPersistentMountDirectory (bool persistentMountDirectory) {
    if (m_persistentMountDirectory == persistentMountDirectory)
        return;

    m_persistentMountDirectory = persistentMountDirectory;
    emit persistentMountDirectoryChanged(persistentMountDirectory);
}

bool Stash::openFileManager() const { return m_openFileManager; }
void Stash::setOpenFileManager (bool openFileManager) {
    if (m_openFileManager == openFileManager)
        return;

    m_openFileManager = openFileManager;
    emit openFileManagerChanged(openFileManager);
}

bool Stash::unmountIdle() const { return m_unmountIdle; }
void Stash::setUnmountIdle (bool unmountIdle) {
    if (m_unmountIdle == unmountIdle)
        return;

    m_unmountIdle = unmountIdle;
    emit unmountIdleChanged(unmountIdle);
}

int Stash::unmountIdleTime() const { return m_unmountIdleTime; }
void Stash::setUnmountIdleTime (int unmountIdleTime) {
    if (m_unmountIdleTime == unmountIdleTime)
        return;

    m_unmountIdleTime = unmountIdleTime;
    emit unmountIdleTimeChanged(unmountIdleTime);
}

QString Stash::encfsConfigFile() const { return m_encfsConfigFile; }
void Stash::setEncfsConfigFile (const QString &encfsConfigFile) {
    QString file = Helpers::pathToAbsolute(encfsConfigFile);
    if (m_encfsConfigFile == file)
        return;

    m_encfsConfigFile = file;
    emit encfsConfigFileChanged(file);
}

QString Stash::additionalMountOptions() const { return m_additionalMountOptions; }
void Stash::setAdditionalMountOptions (const QString &additionalMountOptions) {
    if (m_additionalMountOptions == additionalMountOptions)
        return;

    m_additionalMountOptions = additionalMountOptions;
    emit additionalMountOptionsChanged(additionalMountOptions);
}

QString Stash::preMountCommand() const { return m_preMountCommand; }
void Stash::setPreMountCommand (const QString &preMountCommand) {
    if (m_preMountCommand == preMountCommand)
        return;

    m_preMountCommand = preMountCommand;
    emit preMountCommandChanged(preMountCommand);
}

QString Stash::postMountCommand() const { return m_postMountCommand; }
void Stash::setPostMountCommand (const QString &postMountCommand) {
    if (m_postMountCommand == postMountCommand)
        return;

    m_postMountCommand = postMountCommand;
    emit postMountCommandChanged(postMountCommand);
}

QString Stash::preUnmountCommand() const { return m_preUnmountCommand; }
void Stash::setPreUnmountCommand (const QString &preUnmountCommand) {
    if (m_preUnmountCommand == preUnmountCommand)
        return;

    m_preUnmountCommand = preUnmountCommand;
    emit preUnmountCommandChanged(preUnmountCommand);
}

QString Stash::postUnmountCommand() const { return m_postUnmountCommand; }
void Stash::setPostUnmountCommand (const QString &postUnmountCommand) {
    if (m_postUnmountCommand == postUnmountCommand)
        return;

    m_postUnmountCommand = postUnmountCommand;
    emit postUnmountCommandChanged(postUnmountCommand);
}

QString Stash::keyringName() const { return m_keyringName; }
void Stash::setKeyringName (const QString &keyringName) {
    if (m_keyringName == keyringName)
        return;

    m_keyringName = keyringName;
    emit keyringNameChanged(keyringName);
}

bool Stash::allowNonEmptyMountDirectory() const { return m_allowNonEmptyMountDirectory; }
void Stash::setAllowNonEmptyMountDirectory (bool allowNonEmptyMountDirectory) {
    if (m_allowNonEmptyMountDirectory == allowNonEmptyMountDirectory)
        return;

    m_allowNonEmptyMountDirectory = allowNonEmptyMountDirectory;
    emit allowNonEmptyMountDirectoryChanged(allowNonEmptyMountDirectory);
}

bool Stash::allowRootAccess() const { return m_allowRootAccess; }
void Stash::setAllowRootAccess (bool allowRootAccess) {
    if (m_allowRootAccess == allowRootAccess)
        return;

    m_allowRootAccess = allowRootAccess;
    emit allowRootAccessChanged(allowRootAccess);
}

bool Stash::trayIconIndicateMounted() const { return m_trayIconIndicateMounted; }
void Stash::setTrayIconIndicateMounted (bool trayIconIndicateMounted) {
    if (m_trayIconIndicateMounted == trayIconIndicateMounted)
        return;

    m_trayIconIndicateMounted = trayIconIndicateMounted;
    emit trayIconIndicateMountedChanged(trayIconIndicateMounted);
}

QString Stash::shortName() const { return m_shortName; }
void Stash::setShortName (const QString &shortName) {
    if (m_shortName == shortName)
        return;

    m_shortName = shortName;
    emit shortNameChanged(shortName);
}


// Volatile Properties ============================================================================
bool Stash::isMounted() const { return m_isMounted; }
void Stash::setIsMounted(bool isMounted)
{
    if (m_isMounted == isMounted)
        return;

    m_isMounted = isMounted;
    emit isMountedChanged(isMounted);
}

bool Stash::isAvailable() const { return m_isAvailable; }
void Stash::setIsAvailable(bool isAvailable)
{
    if (m_isAvailable == isAvailable)
        return;

    m_isAvailable = isAvailable;
    emit isAvailableChanged(isAvailable);
}

QString Stash::statusMessage() const { return m_statusMessage; }
bool Stash::isBusy() const { return m_action != Action_Idle; }
void Stash::setBusy(Action action, const QString &message)
{
    if (m_action == action)
        return;

    m_action = action;
    m_statusMessage = message;
    emit isBusyChanged(action != Action_Idle, message);
}

bool Stash::isInitialized() const { return m_isInitialized; }
void Stash::setIsInitialized(bool isInitialized)
{
    if (m_isInitialized == isInitialized)
        return;

    m_isInitialized = isInitialized;
    emit isInitializedChanged(isInitialized);
}

QSharedPointer<Stash> Stash::sharedPointer() const
{
    return m_weakref.toStrongRef();
}

Stash::Action Stash::action() const
{
    return m_action;
}
