#ifndef STASH_H
#define STASH_H

// Qt
#include <QObject>
#include <QSharedPointer>
class QSettings;

// Core
#include "Core/encfs.h"

class Stash : public QObject
{
    Q_OBJECT

public:
    explicit Stash(QObject *parent = nullptr);
    explicit Stash(const QSettings &settings, QObject *parent = nullptr);
    ~Stash();

    enum Action {
        Action_Idle,
        Action_Mounting,
        Action_Unmounting,
        Action_Creating,
        Action_ChangingPassword,
        Action_Deleting
    };
    Q_ENUM(Action)

    void refresh();
    void save(QSettings &settings);
    void load(const QSettings &settings);

    void mount(const QString &password);
    void unmount();
    void create(const QString &password, Encfs::CreateStashFlags flags = 0);
    void changePassword(const QString &oldPassword, const QString &newPassword);
    QString info() const;
    void cancel();
    void delete_();

    bool execCommandLine(const QString &commandLine, bool startDetached = false);

signals:
    void mountFinished(bool success, const QString &message = QLatin1String());
    void unmountFinished(bool success, const QString &message = QLatin1String());
    void createFinished(bool success, const QString &message = QLatin1String());
    void changePasswordFinished(bool success, const QString &message = QLatin1String());
    void deleteFinished(bool success, const QString &message = QLatin1String());

private:
    void buildConnections();
    Encfs &encfs();

private:
    Encfs m_encfs;


// Used the script in in the folder DevTools to autogenerate these
// Saved Properties ===============================================================================
public:
    Q_PROPERTY(QString stashDir READ stashDir WRITE setStashDir NOTIFY stashDirChanged)
    Q_PROPERTY(QString mountDir READ mountDir WRITE setMountDir NOTIFY mountDirChanged)
    Q_PROPERTY(bool mountOnStartup READ mountOnStartup WRITE setMountOnStartup NOTIFY mountOnStartupChanged)
    Q_PROPERTY(bool savePassword READ savePassword WRITE setSavePassword NOTIFY savePasswordChanged)
    Q_PROPERTY(QString stashGroup READ stashGroup WRITE setStashGroup NOTIFY stashGroupChanged)
    Q_PROPERTY(bool showInTray READ showInTray WRITE setShowInTray NOTIFY showInTrayChanged)
    Q_PROPERTY(bool mountWhenAvailable READ mountWhenAvailable WRITE setMountWhenAvailable NOTIFY mountWhenAvailableChanged)
    Q_PROPERTY(bool allowOtherUsers READ allowOtherUsers WRITE setAllowOtherUsers NOTIFY allowOtherUsersChanged)
    Q_PROPERTY(bool persistentMountDirectory READ persistentMountDirectory WRITE setPersistentMountDirectory NOTIFY persistentMountDirectoryChanged)
    Q_PROPERTY(bool openFileManager READ openFileManager WRITE setOpenFileManager NOTIFY openFileManagerChanged)
    Q_PROPERTY(bool unmountIdle READ unmountIdle WRITE setUnmountIdle NOTIFY unmountIdleChanged)
    Q_PROPERTY(int unmountIdleTime READ unmountIdleTime WRITE setUnmountIdleTime NOTIFY unmountIdleTimeChanged)
    Q_PROPERTY(QString encfsConfigFile READ encfsConfigFile WRITE setEncfsConfigFile NOTIFY encfsConfigFileChanged)
    Q_PROPERTY(QString additionalMountOptions READ additionalMountOptions WRITE setAdditionalMountOptions NOTIFY additionalMountOptionsChanged)
    Q_PROPERTY(QString preMountCommand READ preMountCommand WRITE setPreMountCommand NOTIFY preMountCommandChanged)
    Q_PROPERTY(QString postMountCommand READ postMountCommand WRITE setPostMountCommand NOTIFY postMountCommandChanged)
    Q_PROPERTY(QString preUnmountCommand READ preUnmountCommand WRITE setPreUnmountCommand NOTIFY preUnmountCommandChanged)
    Q_PROPERTY(QString postUnmountCommand READ postUnmountCommand WRITE setPostUnmountCommand NOTIFY postUnmountCommandChanged)
    Q_PROPERTY(QString keyringName READ keyringName WRITE setKeyringName NOTIFY keyringNameChanged)
    Q_PROPERTY(bool allowNonEmptyMountDirectory READ allowNonEmptyMountDirectory WRITE setAllowNonEmptyMountDirectory NOTIFY allowNonEmptyMountDirectoryChanged)
    Q_PROPERTY(bool allowRootAccess READ allowRootAccess WRITE setAllowRootAccess NOTIFY allowRootAccessChanged)
    Q_PROPERTY(bool trayIconIndicateMounted READ trayIconIndicateMounted WRITE setTrayIconIndicateMounted NOTIFY trayIconIndicateMountedChanged)
    Q_PROPERTY(QString shortName READ shortName WRITE setShortName NOTIFY shortNameChanged)

    QString stashDir() const;
    QString mountDir() const;
    bool mountOnStartup() const;
    bool savePassword() const;
    QString stashGroup() const;
    bool showInTray() const;
    bool mountWhenAvailable() const;
    bool allowOtherUsers() const;
    bool persistentMountDirectory() const;
    bool openFileManager() const;
    bool unmountIdle() const;
    int unmountIdleTime() const;
    QString encfsConfigFile() const;
    QString additionalMountOptions() const;
    QString preMountCommand() const;
    QString postMountCommand() const;
    QString preUnmountCommand() const;
    QString postUnmountCommand() const;
    QString keyringName() const;
    bool allowNonEmptyMountDirectory() const;
    bool allowRootAccess() const;
    bool trayIconIndicateMounted() const;
    QString shortName() const;

signals:
    void stashDirChanged (const QString &stashDir);
    void mountDirChanged (const QString &mountDir);
    void mountOnStartupChanged (bool mountOnStartup);
    void savePasswordChanged (bool savePassword);
    void stashGroupChanged (const QString &stashGroup);
    void showInTrayChanged (bool showInTray);
    void mountWhenAvailableChanged (bool mountWhenAvailable);
    void allowOtherUsersChanged (bool allowOtherUsers);
    void persistentMountDirectoryChanged (bool persistentMountDirectory);
    void openFileManagerChanged (bool openFileManager);
    void unmountIdleChanged (bool unmountIdle);
    void unmountIdleTimeChanged (int unmountIdleTime);
    void encfsConfigFileChanged (const QString &encfsConfigFile);
    void additionalMountOptionsChanged (const QString &additionalMountOptions);
    void preMountCommandChanged (const QString &preMountCommand);
    void postMountCommandChanged (const QString &postMountCommand);
    void preUnmountCommandChanged (const QString &preUnmountCommand);
    void postUnmountCommandChanged (const QString &postUnmountCommand);
    void keyringNameChanged (const QString &keyringName);
    void allowNonEmptyMountDirectoryChanged (bool allowNonEmptyMountDirectory);
    void allowRootAccessChanged (bool allowRootAccess);
    void trayIconIndicateMountedChanged(bool trayIconIndicateMounted);
    void shortNameChanged(const QString &shortName);

public slots:
    void setStashDir (const QString &stashDir);
    void setMountDir (const QString &mountDir);
    void setMountOnStartup (bool mountOnStartup);
    void setSavePassword (bool savePassword);
    void setStashGroup (const QString &stashGroup);
    void setShowInTray (bool showInTray);
    void setMountWhenAvailable (bool mountWhenAvailable);
    void setAllowOtherUsers (bool allowOtherUsers);
    void setPersistentMountDirectory (bool persistentMountDirectory);
    void setOpenFileManager (bool openFileManager);
    void setUnmountIdle (bool unmountIdle);
    void setUnmountIdleTime (int unmountIdleTime);
    void setEncfsConfigFile (const QString &encfsConfigFile);
    void setAdditionalMountOptions (const QString &additionalMountOptions);
    void setPreMountCommand (const QString &preMountCommand);
    void setPostMountCommand (const QString &postMountCommand);
    void setPreUnmountCommand (const QString &preUnmountCommand);
    void setPostUnmountCommand (const QString &postUnmountCommand);
    void setKeyringName (const QString &keyringName);
    void setAllowNonEmptyMountDirectory (bool allowNonEmptyMountDirectory);
    void setAllowRootAccess (bool allowRootAccess);
    void setTrayIconIndicateMounted(bool trayIconIndicateMounted);
    void setShortName(const QString &shortName);

private:
    QString m_stashDir;
    QString m_mountDir;
    bool m_mountOnStartup;
    bool m_savePassword;
    QString m_stashGroup;
    bool m_showInTray;
    bool m_mountWhenAvailable;
    bool m_allowOtherUsers;
    bool m_persistentMountDirectory;
    bool m_openFileManager;
    bool m_unmountIdle;
    int m_unmountIdleTime;
    QString m_encfsConfigFile;
    QString m_additionalMountOptions;
    QString m_preMountCommand;
    QString m_postMountCommand;
    QString m_preUnmountCommand;
    QString m_postUnmountCommand;
    QString m_keyringName;
    bool m_allowNonEmptyMountDirectory;
    bool m_allowRootAccess;
    bool m_trayIconIndicateMounted;
    QString m_shortName;


// Volatile Properties ============================================================================
public:
    Q_PROPERTY(bool isMounted READ isMounted WRITE setIsMounted NOTIFY isMountedChanged)
    Q_PROPERTY(bool isAvailable READ isAvailable WRITE setIsAvailable NOTIFY isAvailableChanged)
    Q_PROPERTY(bool isBusy READ isBusy NOTIFY isBusyChanged)
    Q_PROPERTY(bool isInitialized READ isInitialized WRITE setIsInitialized NOTIFY isInitializedChanged)
    Q_PROPERTY(Action action READ action)

    bool isMounted() const;
    bool isAvailable() const;
    bool isBusy() const;
    QString statusMessage() const;
    bool isInitialized() const;
    Action action() const;

public slots:
    void setIsInitialized(bool isInitialized);


private:
    QSharedPointer<Stash> sharedPointer() const;

signals:
    void isMountedChanged(bool isMounted);
    void isAvailableChanged(bool isaAvailable);
    void isBusyChanged(bool isBusy, const QString &message = QLatin1String());
    void isInitializedChanged(bool isInitialized);
    void __deleteFinished(bool success, const QString &message = QLatin1String());

private slots:
    void setIsMounted(bool isMounted);
    void setIsAvailable(bool isaAvailable);
    void setBusy(Stash::Action action, const QString &message = QLatin1String());

private:
    bool m_isMounted;
    bool m_isAvailable;
    QString m_statusMessage;
    bool m_isInitialized;
    bool m_passwordInKeyring;
    Action m_action;
    // is set by the StashManager when added
    QWeakPointer<Stash> m_weakref;

public:
    Encfs::CreateStashFlags m_createFlags;

    friend class StashManager;
};

#endif // STASH_H
